Option Explicit

' Miscellaneous error-handling functions.
' Among other things, this module implements a JScript-like try/catch facility 
' for VBScript.  See ErhCls, ErhBegin, and EndErh for details.
' Requires Trace.vbs.


Public Const errnoOK = 0
Public Const errnoAssert = 16601
Public Const errnoInvalidKloc = 16602
Public Const errnoTimeout = 16603
Public Const errnoUnwindError = 16604
Public Const errnoInvalidRid = 16605

Public wavError
Set wavError = skapi.WavNew("Error.wav")


' Raise an error if the specified Boolean condition f is not true.  
' sz is a string describing the condition for error-reporting purposes.
Public Function Assert(f, sz)
	
	If Not f Then
		Call Err.Raise(errnoAssert, "Assert", sz)
	End If
	
End Function

' Check to see if the given timer has exceeded the timeout specified by 
' csecTimeout (in seconds).  If so, raise an error using szSrc as the 
' "Source" parameter to the error message.
Public Function CheckTimeout(timer, csecTimeout, szSrc)
	
	If timer.cmsec > csecTimeout * 1000 Then
		Call RaiseTimeout(szSrc)
	End If
	
End Function

' Raise a timeout error using szSrc as the "Source" parameter to the error message. 
Public Function RaiseTimeout(szSrc)
	
	Call Err.Raise(errnoTimeout, szSrc, "Timeout")
	
End Function

' Check skapi.arc.  If it's not arcSuccess, raise an error.
Public Function CheckArc()
	
	Dim arc : arc = skapi.arc
	If arc <> arcSuccess Then
		Call RaiseCmr(SzFromArc(arc))
	End If
	
End Function

' Convert a skapi.arc error number into a human-readable string.
Public Function SzFromArc(arc)
	
	Select Case arc
	Case arcTooBusy
		SzFromArc = "You're too busy."
	Case arcCantMoveToObject
		SzFromArc = "Unable to move to object."
	Case Else
		SzFromArc = "arc = " + CStr(arc)
	End Select
	
End Function

' Raise a generic error.  sz is the text of the error message.
Public Function RaiseCmr(sz)
	
	Call Err.Raise(-1, "RaiseCmr", sz)
	
End Function


' Error Handler object.
' Use this to emulate JScript's try/catch feature as follows:
' 
'	Dim erh : Set erh = ErhBegin()
'	On Error Resume Next
'	
'	' Some code that might raise an error (the 'try' clause).
'	' (Best if this is a single statement such as a subroutine call.)
'	
'	Call EndErh(erh)
'	On Error Goto 0
'	If erh.errno <> errnoOK Then
'		
'		' Handle the error (the 'catch' clause).
'		
'	End If
' 
Class ErhCls
	
	Public trast		' Trace state for unwinding the Trace stack.
	Public iobjUnw		' Unwind index for calling unwind handlers.
	Public errno		' The error number, or errnoOK if no error.
	Public szSource		' A text string indicating the source of the error.
	Public szDesc		' The text of the error message.
	
'	Return a textual error message representing this error.
	Public Property Get sz()

		If errno = errnoOK Then
			sz = ""
		ElseIf errno = -1 Then
			sz = szDesc
		Else
			sz = szSource + ": " + szDesc + " (error " + CStr(errno) + ")"
		End If
		
	End Property
	
'	Re-throw the error to be caught by a higher-level handler.
	Public Function PassItOn()
		
		If errno <> 0 Then
			Call Err.Raise(errno, "~!" + szSource, szDesc)
		End If
		
	End Function
	
End Class

' Initialize and return an error handler object.
' Must be paired with EndErh as in the above example.
Public Function ErhBegin()
	
	Dim erh : Set erh = New ErhCls
	erh.trast = TrastGetTrace()
	erh.iobjUnw = iobjUnwind
	Set ErhBegin = erh
	
End Function

' Capture error parameters and clean up.
' Must be paired with ErhBegin as in the above example.
Public Function EndErh(erh)
	
'	Save away the error info.
	erh.errno = Err.Number
	erh.szSource = Err.Source
	erh.szDesc = Err.Description
	
'	Was there in fact an error?
	If erh.errno <> errnoOK Then
	'	Yes.
		If FPrefixMatchSz(erh.szSource, "~!") Then
		'	We're re-catching an error re-thrown by erh.PassItOn.
		'	Don't re-report the error in this case.
			erh.szSource = Right(erh.szSource, Len(erh.szSource) - 2)
		Else
		'	First catch; play the error tone.
			Call wavError.Play()
		'	Log the error to the Trace log.
			Call TraceLineAlways(">>> " + erh.sz)
			Call TraceLineAlways("skapi.plig = " + CStr(skapi.plig))
		End If
	'	Unwind the Trace stack.
		Call TraceUnwind(erh.trast)
	'	Call any registered Unwind handlers to give unwound code a chance to clean up.
		Call Unwind(erh)
	Else
	'	Normal case (no error).  Verify proper ErhBegin/EndErh pairing.
		If TrastGetTrace() <> erh.trast Or _
		  iobjUnwind <> erh.iobjUnw Then
		'	Log usage error.
			erh.errno = errnoUnwindError
			erh.szSource = "EndErh"
			erh.szDesc = "Unwind error"
			Call wavError.Play()
			Call TraceLineAlways(">>> " + erh.sz)
			Call TraceLineAlways("TrastGetTrace() = " + CStr(TrastGetTrace()) + _
			  ", erh.trast = " + CStr(erh.trast))
			Call TraceLineAlways("iobjUnwind = " + CStr(iobjUnwind) + _
			  ", erh.iobjUnw = " + CStr(erh.iobjUnw))
		End If
	End If
	
End Function


' Register an unwind handler to be called in case of error.
Public Function PushUnwind(obj)
	
	Call Assert(iobjUnwind <= UBound(rgobjUnwind), "unwind overflow")
	Set rgobjUnwind(iobjUnwind) = obj
	iobjUnwind = iobjUnwind + 1
	
End Function

' Unregister an unwind handler.
Public Function PopUnwind(obj)
	
	Call Assert(iobjUnwind > 0, "unwind underflow")
	iobjUnwind = iobjUnwind - 1
	Call Assert(rgobjUnwind(iobjUnwind) Is obj, "unwind mismatch")
	Set rgobjUnwind(iobjUnwind) = Nothing
	
End Function


'----------
' Privates
'----------

Private rgobjUnwind(30)		' Stack of registered unwind handlers.
Private iobjUnwind			' Unwind stack pointer.
iobjUnwind = 0

' Pop and call undwind handlers back to the appropriate level.
Private Function Unwind(erh)
	
	Do While iobjUnwind > erh.iobjUnw
		iobjUnwind = iobjUnwind - 1
		Call rgobjUnwind(iobjUnwind).Unwind(erh)
		Set rgobjUnwind(iobjUnwind) = Nothing
	Loop
	
End Function


