Option Explicit

' This module provides functions for outputting an execution trace to the 
' debug log.  The trace logs procedure entries and exits using indentation 
' to indicate procedure nesting and square brackets to match up corresponding 
' entries and exits.
' 
' Tracing may be enabled or disabled globally by calling Trace(Nothing) or 
' NoTrace(Nothing), per module by calling Trace(SomeModule) or NoTrace(SomeModule), 
' or per function using the tef parameter to TraceEnterEx.


'------------------
' Public functions
'------------------

' Enable tracing of the module specified by obj.  If obj is Nothing, 
' enable tracing globally.
Public Function Trace(obj)
	If obj Is Nothing Then
		fTrace = True
	Else
		obj.tef = tefEnable
	End If
End Function

' Disable tracing of the module specified by obj.  If obj is Nothing, 
' disable tracing globally.
Public Function NoTrace(obj)
	If obj Is Nothing Then
		fTrace = False
	Else
		obj.tef = tefDisable
	End If
End Function

' Return True is tracing is enabled at the current level of procedure nesting.
Public Function FTraceEnabled()
	
	If fTrace Then
		FTraceEnabled = ((rgtef(clevelIndent) And tefEnable) <> 0 Or fDebugTrace)
	Else
		FTraceEnabled = False
	End If
	
End Function

' Log a procedure entry.  Call this once on entry to each function you want 
' logged.  Good practice is to make it the first line of the function body.  
' szFunc is the name of the function; szArgs is a string representation of the 
' argument lists (without parentheses).  You're free to encode as many or as 
' few arguments as you like in szArgs; for no arguments, pass the empty string.
' 
' Each call to TraceEnter must be matched by a call to TraceExit.  Beware of 
' doing an Exit Sub or Exit Function out of the middle of a traced function, 
' or you'll trigger an Assert in TraceExit.
Public Function TraceEnter(szFunc, szArgs)
	
	If Not fTrace Then Exit Function
	
'	Pass down the current Enable/Disable flag.
	Call TraceEnterEx(szFunc, szArgs, rgtef(clevelIndent) And tefEnable)
	
End Function

' Trace Enter Flags:
Public Const tefDisable		= &H00	' Tracing of subroutines is disabled (the default).
Public Const tefEnable		= &H01	' Tracing of subroutines is enabled.
Public Const tefNonEmpty	= &H02	' Suppress call if no subroutines are called.
Public Const tefFull		= &H04	' Full trace from here down.

' Log a procedure entry as in TraceEnter, with options as specified by tef.
Public Function TraceEnterEx(szFunc, szArgs, tef)
	
	If Not fTrace Then Exit Function
	
	Call Assert(clevelIndent + 1 <= UBound(rgtef), "clevelIndent overflow")
	rgszFunc(clevelIndent + 1) = szFunc
	rgszArgs(clevelIndent + 1) = szArgs
	If rgtef(clevelIndent) = tefFull Then
		rgtef(clevelIndent + 1) = tefFull
	Else
		rgtef(clevelIndent + 1) = tef
	End If
	
'	If either me or my caller is enabled, show me.
	If ((rgtef(clevelIndent) Or rgtef(clevelIndent + 1)) And tefEnable) <> 0 Or _
	  fDebugTrace Then
		If Len(szTraceLine) > 0 Then Call TraceEndLine()
		Call TraceStartLine()
		Call TraceSz("[ " + rgszFunc(clevelIndent + 1))
		If szArgs <> "" Then Call TraceSz("(" + szArgs + ")")
		If rgtef(clevelIndent + 1) = tefFull Then Call TraceEndLine()
	End If
	
	clevelIndent = clevelIndent + 1
	
End Function

' Log a procedure exit.  Call this once on exiting each function you want 
' logged.  Good practice is to make it the last line of the function body, 
' and to avoid using Exit Function or Exit Sub within that function.  szFunc 
' is the name of the function; szResult is a string representation of the 
' function's return value.  For no return value, pass the empty string.
' 
' Each call to TraceExit must match a prior call to TraceEnter or TraceEnterEx.
Public Function TraceExit(szFunc, szResult)
	
	If Not fTrace Then Exit Function

	If szFunc <> rgszFunc(clevelIndent) Then
		Call Assert(False, _
		  "TraceEnter/Exit mismatch: " + _
		  SzQuote(szFunc) + " <> " + SzQuote(rgszFunc(clevelIndent)))
	End If
	Call Assert(clevelIndent > 0, "clevelIndent underflow")
	clevelIndent = clevelIndent - 1
	
'	If either me or my caller is enabled, show me.
	If ((rgtef(clevelIndent) Or rgtef(clevelIndent + 1)) And tefEnable) <> 0 Or _
	  fDebugTrace Then
		If Len(szTraceLine) > 0 And _
		  rgtef(clevelIndent + 1) <> tefFull Then
			If (rgtef(clevelIndent + 1) And tefNonEmpty) <> 0 Then
			'	Entry line is still pending; discard it.
				Call TraceDiscardLine()
			Else
			'	Show exit on same line if nothing intervened.
				If szResult <> "" Then Call TraceSz(": " + szResult)
				Call TraceSz(" ]")
				Call TraceEndLine()
			End If
		Else
		'	Show exit on a new line.
			If Len(szTraceLine) > 0 Then Call TraceEndLine()
			Call TraceStartLine()
			Call TraceSz("] " + rgszFunc(clevelIndent + 1))
			If szResult <> "" Then Call TraceSz(": " + szResult)
			Call TraceEndLine()
		End If
	End If
	
End Function

' Log an arbitrary line of text to the execution trace.  sz is the text to log.  
' Do not include a terminating CR/LF; TraceLine supplies that automatically.  
' Call TraceLine as many or as few times as you like within a traced function; 
' the displayed lines will be indented automatically according to the function 
' nesting.
Public Function TraceLine(sz)
	
	If Not FTraceEnabled() Then Exit Function
	
	Call TraceLineAlways(sz)
	
End Function

' Log an arbitrary line of text as in TraceLine, regardless of the current 
' enable/disable state, so that the line always appears in the trace log.
Public Function TraceLineAlways(sz)
	
	If Len(szTraceLine) > 0 Then Call TraceEndLine()
	Call TraceStartLine()
	Call TraceSz(sz)
	Call TraceEndLine()
	
End Function

' Log an input event to the execution trace.  sz is a string representation of 
' the event.  Note that "input" here refers to keystrokes and mouse clicks 
' your script generates.  So it's input from AC's point of view, but output 
' from your point of view.
' 
' This differs from TraceLine in that input events all line up in a column 
' to the right, rather than following the function nesting indentation.
Public Function TraceInput(sz)
	
	If Not FTraceEnabled() Then Exit Function
	
'	ichInput is the zero-origin character position where input events are aligned.
	Const ichInput = 45
	
	Call TraceSeekIch(ichInput)
	Call TraceSz(sz)
	Call TraceEndLine()
	
End Function

' Return a string representation of the given object suitable for use in the 
' szArgs parameter of TraceEnter or the szResult parameter of TraceExit.
' Assumes the object implements an szTrace method.
Public Function SzTraceObj(obj)
	If obj Is Nothing Then
		SzTraceObj = "Nothing"
	Else
		SzTraceObj = obj.szTrace
	End If
End Function


' Return the current trace nesting state.
' Used by Error.vbs for error recovery.
Public Function TrastGetTrace()
	
	TrastGetTrace = clevelIndent
	
End Function

' Unwind the trace to the specified state by simulating procedure exits.  
' Used by Error.vbs for error recovery.
Public Function TraceUnwind(trast)
	
	Dim clevel : clevel = trast
	
	If clevelIndent > clevel Then
	'	Show everybody on the Unwind stack.
		rgtef(clevelIndent) = rgtef(clevelIndent) Or tefEnable
		If Len(szTraceLine) > 0 Then Call TraceEndLine()
		Do While clevelIndent > clevel And clevelIndent > 0
			clevelIndent = clevelIndent - 1
			Call TraceStartLine()
			Call TraceSz("] " + "unwind " + rgszFunc(clevelIndent + 1))
			If rgszArgs(clevelIndent + 1) <> "" Then
				Call TraceSz("(" + rgszArgs(clevelIndent + 1) + ")")
			End If
			Call TraceEndLine()
			rgtef(clevelIndent) = rgtef(clevelIndent) Or tefEnable
		Loop
	End If
	
End Function


'----------
' Privates
'----------

Private fTrace					' Global enable/disable flag.
fTrace = True

' Set this to True for verbose output when debugging the Trace module itself.
Private Const fDebugTrace = False

Private rgszFunc(50)			' Stack of procedure names.
Private rgszArgs(50)			' Stack of procedure arguments.
Private rgtef(50)				' Stack of TraceEntry flags.
Private clevelIndent			' Call stack indentation level.

rgtef(0) = tefEnable			' Enable tracing of global init code.

Private szTraceLine				' Line buffer for trace output.


' Start a line of trace output, including indentation.
Private Function TraceStartLine()
	
	Call Assert(Len(szTraceLine) = 0, "Len(szTraceLine) > 0")
	If fDebugTrace Then
		Dim ilevel
		For ilevel = 1 To clevelIndent
			If (rgtef(ilevel) And tefEnable) <> 0 Then
				Call TraceSz("+ ")
			Else
				Call TraceSz("- ")
			End If
		Next
	Else
		Call TraceSz(Space(clevelIndent * 2))
	End If
	
End Function

' Position the output buffer to the specified column position, starting 
' a new line if necessary.
Private Function TraceSeekIch(ich)

	If Len(szTraceLine) >= ich Then Call TraceEndLine()
	
	Call TraceSz(Space(ich - Len(szTraceLine)))
	
End Function

' Append the given string to the output buffer.
Private Function TraceSz(sz)

	szTraceLine = szTraceLine + sz
	
End Function

' Output the contents of the line buffer.
Private Function TraceEndLine()
	
	Call skapi.OutputLine(szTraceLine, opmDebugLog)
	szTraceLine = ""
	
End Function

' Discard the contents of the line buffer.
Private Function TraceDiscardLine()
	
	szTraceLine = ""
	
End Function



