Option Explicit
Class WayptDoorCls
' Door-opening waypoint class.
' Supports automatic unlocking of locked doors (if you have the proper key.)


'---------------------------------------
' Type-specific properties and methods:
'---------------------------------------

' Return the location of the door itself (as distinct from maplocIn and maplocOut).
Public Property Get maploc()
	Set maploc = maplocMbr
End Property

' Set the location of the door itself (as distinct from maplocIn and maplocOut).
Public Property Set maploc(maplocA)
	
	Set maplocMbr = maplocA
	
	If maplocMbr.head >= 0 Then
	'	Calculate maplocIn and maplocOut from the door's heading.
	'	(A door's heading is the direction you face to go through it.)
		Dim dlat, dlng
		dlat = distDoor * Cos(maplocMbr.headRad)
		dlng = distDoor * Sin(maplocMbr.headRad)
		Call TraceLine("dlat = " + FormatNumber(dlat, 3) + ", " + _
		  "dlng = " + FormatNumber(dlng, 3))
		Set maplocInMbr = skapi.MaplocFromLatLng(maplocMbr.lat - dlat, _
		  maplocMbr.lng - dlng, maplocMbr.z, maplocMbr.head)
		Set maplocOutMbr = skapi.MaplocFromLatLng(maplocMbr.lat + dlat, _
		  maplocMbr.lng + dlng, maplocMbr.z, maplocMbr.head)
	Else
	'	Door's heading is unknown.  Defer calculation of maplocIn and maplocOut 
	'	until we get there.
		Set maplocInMbr = maplocMbr
		Set maplocOutMbr = maplocMbr
	End If
	
End Property

Public szKey			' The name of the key (if any) that unlocks it.

' Initialize this waypoint from the given parameters.
Public Function Init(maplocA, szKeyA)
	Call TraceEnterEx("WayptDoor.Init", _
	  SzTraceMaploc(maplocA) + ", " + SzQuote(szKeyA), Xnav.tef)
	
	Set maploc = maplocA
	szKey = szKeyA
	
	Call TraceExit("WayptDoor.Init", "")
End Function


'----------------------------------
' Required properties and methods:
'----------------------------------

' The XML tag name of this waypoint type.
Public Property Get szTag()
	szTag = "Door"
End Property

' True if this leg should be walked, false if it should be run.
' Xnav handles initialization of this property, but each waypoint class must define it.
Public fWalk

' The maploc Xnav should steer toward.
' This will be a point on the approach side of the door.
Public Property Get maplocIn()
	Set maplocIn = maplocInMbr
End Property

' The maploc we're at after the arrival action has taken place.
' I.e. after we've gone through the door.
Public Property Get maplocOut()
	Set maplocOut = maplocOutMbr
End Property

' The arrival distance of this waypoint in MU, i.e. how accurately we need to approach it 
' before performing the arrival action (if any) and passing on to the next waypoint.
Public Property Get distArrive()
	distArrive = 2 * Xnav.distMeter	' Open door when we get within 2 meters.
End Property

' The estimated time in seconds needed for this waypoint's arrival action.
' Estimate five seconds on average to open the door and go through it.
Public Property Get csecArrive()
	csecArrive = 5
End Property

' Perform this waypoint's arrival action.
Public Function Arrive()
	Call TraceEnterEx("WayptDoor.Arrive", SzTraceMaploc(maploc), Xnav.tef)
	Call PushUnwind(Me)
	
'	Stop moving while we find the door and open it.
	Xnav.mstate = Xnav.mstateStop
	
'	Locate the Door object.
	Set acoDoor = AcoFindDoor()
	If acoDoor Is Nothing Then Call RaiseCmr("Can't find door.")
	
'	If door's head is unknown, figure it out now.
	If maploc.head < 0 Then
		Set maploc = MaplocDoorFromAco(acoDoor, skapi.maplocCur)
	End If
	
'	Try up to five times to go through.
	Const ctry = 5
	Dim itry, erh
	For itry = 0 To ctry - 1
	'	Make sure the door is unlocked and open.
		
	'	Error-handling boilerplate.
		Set erh = ErhBegin()
		On Error Resume Next
		
		Call EnsureOpen()
		
	'	Error-handling boilerplate.
		Call EndErh(erh)
		On Error Goto 0
		
		If erh.errno = errnoOK Then
			
			Call skapi.WaitEvent()
			Call skapi.AddHandler(evidOnNavStop, Me)
			
		'	Approach this side of the door.  Our OnNavStop handler will take care of 
		'	getting us through the door to the other side.
			Call TraceLine("skapi.GoToMaploc(" + SzTraceMaploc(maplocIn) + ", " + _
			  FormatNumber(distDoor/2, 3) + ", " + CStr(True) + ")")
			Call skapi.GoToMaploc(maplocIn, distDoor/2, True, False)
			Xnav.mstate = Xnav.mstateUnknown
			Call TraceLine("state = stateThisSide")
			state = stateThisSide
			
		'	Loop until we make it through or time out trying.
			fBreak = False
			Do Until fBreak
				If Not Pauser.fRunning Or Xnav.fStopNav Then
					nscBreak = nscCanceled
					Exit Do
				End If
				Call skapi.WaitEvent(500, wemSpecific, evidOnNavStop)
			Loop
			
			Call skapi.RemoveHandler(evidOnNavStop, Me)
			
			If nscBreak = nscArrived And state = stateThrough Then Exit For
			If nscBreak = nscCanceled Then Exit For
			
		ElseIf erh.errno <> errnoTimeout Then
			Call erh.PassItOn()
		End If
		
	'	Didn't make it; try opening it again.
	Next
	If itry = ctry Then
		Call RaiseCmr("Can't get through " + acoDoor.szName + " at " + acoDoor.maploc.sz(3) + ".")
	End If
	
	Set acoDoor = Nothing
	
	Call PopUnwind(Me)
	Call TraceExit("WayptDoor.Arrive", "")
End Function

' Generate a new waypoint representing the reversal of this waypoint.
' Used by Xroute.XrouteReverse.
Public Function WayptReverse()
	Call TraceEnterEx("WayptDoor.WayptReverse", SzTraceMaploc(maploc), Xnav.tef)
	
'	Reverse the door's heading, thereby swapping maplocIn and maplocOut.
	Set WayptReverse = WayptDoorNew()
	Call WayptReverse.Init(Xnav.MaplocReverseHead(maploc), szKey)
	
	Call TraceExit("WayptDoor.WayptReverse", "")
End Function

' Initialize this waypoint from skapi.maplocCur and, if relevant, skapi.acoSelected.
' Used by XrouteEdit.
Public Function FromMaplocCur()
	Call TraceEnterEx("WayptDoor.FromMaplocCur", "", Xnav.tef)
	
'	Make sure we have a door selected.
	Dim acoSel : Set acoSel = skapi.acoSelected
	If acoSel Is Nothing Then Call RaiseCmr("Please select a door.")
	If Not FAcoIsDoor(acoSel) Then Call RaiseCmr("Please select a door.")
	
	Call Init(MaplocDoorFromAco(acoSel, skapi.maplocCur), "")
	
	Call TraceExit("WayptDoor.FromMaplocCur", "")
End Function

' Initialize this waypoint from an XML element.
Public Function FromXml(elem)
	Call TraceEnterEx("WayptDoor.FromXml", SzQuote(elem.xml), Xnav.tef)
	
	Dim maploc : Set maploc = Xnav.MaplocOfElem(elem)
	
	Call Init(maploc, SzGetAttr(elem, "szKey"))
	
	Call TraceExit("WayptDoor.FromXml", "")
End Function

' Save type-specific waypoint parameters to the given XML element.
' If fVerbose is True, all waypoint parameters should be made explicit in the XML.  
' (This is needed by XrouteEdit.)  If fVerbose is False, parameters with default 
' values may be omitted.
Public Function ToXml(elem, fVerbose)
	Call TraceEnterEx("WayptDoor.ToXml", CStr(fVerbose), Xnav.tef)
	
	Call Xnav.SetElemMaploc(elem, maploc, True, True)
	
	If szKey <> "" Or fVerbose Then
		Call SetSzAttr(elem, "szKey", szKey)
	End If
	
	Call TraceExit("WayptDoor.ToXml", "")
End Function


'-----------------
' Event handlers:
'-----------------

Public Function OnNavStop(nsc)
	Call TraceEnterEx("WayptDoor.OnNavStop", CStr(nsc), Xnav.tef)
	
	nscBreak = nsc
	If nsc = nscArrived Then
		Select Case state
		Case stateThisSide
		'	Made it to maplocIn.
		'	If we're more than 45 degrees off the correct heading, stop and turn in place.
			If Abs(Xnav.DheadNormalize(skapi.maplocCur.head - maploc.head)) > 45 Then
				Xnav.mstate = Xnav.mstateStop
			End If
		'	Now start moving through the door toward maplocOut.
			Call TraceLine("skapi.GoToMaploc(" + SzTraceMaploc(maplocOut) + ", " + _
			  FormatNumber(distDoor/2, 3) + ", " + CStr(True) + ")")
			Call skapi.GoToMaploc(maplocOut, distDoor/2, True, False)
			Xnav.mstate = Xnav.mstateUnknown
			Call TraceLine("state = stateOtherSide")
			state = stateOtherSide
		Case stateOtherSide
		'	Made it to maplocOut; we're done.
			Call TraceLine("state = stateThrough")
			state = stateThrough
			fBreak = True
		End Select
	Else
	'	Timed out somewhere along the way.
	'	Back off and retry from the top in WayptDoor.Arrive.
		fBreak = True
	End If
	
	Call TraceExit("WayptDoor.OnNavStop", "")
End Function

Public Function OnSetFlagOther(aco, iflag, fOn)
	Call TraceEnterEx("WayptDoor.OnSetFlagOther", _
	  SzTraceAco(aco) + ", " + CStr(iflag) + ", " + CStr(fOn), Xnav.tef)
	
	If aco Is acoDoor And iflag = 3 Then
		fLocked = fOn
	End If
	
	Call TraceExit("WayptDoor.OnSetFlagOther", "")
End Function

Public Function OnTipMessage(szMsg)
	Call TraceEnterEx("WayptDoor.OnTipMessage", SzQuote(szMsg), Xnav.tef)
	
	If FEqSzI(szMsg, "The door is locked!") Then
		fLocked = True
	End If
	
	Call TraceExit("WayptDoor.OnTipMessage", "")
End Function


'----------------
' Unwind handler
'----------------

' Remove event handlers and such in case of error.
Public Function Unwind(erh)
	Call TraceEnterEx("WayptDoor.Unwind", "", tefEnable)
	
	Call skapi.RemoveHandler(evidOnNavStop, Me)
	Call skapi.RemoveHandler(evidOnTipMessage, Me)
	Set acoDoor = Nothing
	
	Call TraceExit("WayptDoor.Unwind", "")
End Function


'-----------
' Privates:
'-----------

Private maplocMbr	' The location of the door itself.

Private Property Get distDoor()
	distDoor = 1.0 * Xnav.distMeter
End Property
Private maplocInMbr, maplocOutMbr

Private acoDoor
Private fLocked

Private Property Get stateThisSide()
	stateThisSide = 0
End Property
Private Property Get stateOtherSide()
	stateOtherSide = 1
End Property
Private Property Get stateThrough()
	stateThrough = 2
End Property
Private state
Private fBreak, nscBreak

Private Sub Class_Initialize()
	
	Set maplocMbr = Nothing
	Set maplocInMbr = Nothing
	Set maplocOutMbr = Nothing
	
End Sub

' Given a door and an approach direction, return the door's maploc with 
' the appropriate heading for that approach.
Private Function MaplocDoorFromAco(aco, maplocApproach)
	Call TraceEnter("WayptDoor.MaplocDoorFromAco", _
	  SzTraceAco(aco) + ", " + SzTraceMaploc(maplocApproach))
	
	Dim maplocDoor : Set maplocDoor = aco.maploc
	Dim headToDoor : headToDoor = maplocApproach.HeadToMaploc(maplocDoor)
	If Abs(Xnav.DheadNormalize(headToDoor - maplocDoor.head)) > 90 Then
	'	We're approaching the door from the back side (opposite maploc.head).
	'	Flip the door's head so maplocIn and maplocOut come out right.
		Set maplocDoor = Xnav.MaplocReverseHead(maplocDoor)
	End If
	
	Set MaplocDoorFromAco = maplocDoor
	
	Call TraceExit("WayptDoor.MaplocDoorFromAco", SzTraceMaploc(MaplocDoorFromAco))
End Function

' Look for a door near the expected coordinates.
Private Function AcoFindDoor()
	Call TraceEnter("WayptDoor.AcoFindDoor", SzTraceMaploc(maploc))
	
	Dim acf : Set acf = skapi.AcfNew()
	Set acf.maploc = skapi.MaplocFromLatLng(maploc.lat, maploc.lng, _
	  skapi.maplocCur.z)
	acf.oty = otyDoor
	acf.distMax = 0.1
	
	Set AcoFindDoor = acf.AcoGetSorted(skapi.sortbyDist)
	
	Call TraceExit("WayptDoor.AcoFindDoor", SzTraceAco(AcoFindDoor))
End Function


Private Function EnsureOpen()
	Call TraceEnterEx("WayptDoor.EnsureOpen", SzTraceMaploc(maploc), _
	  tefEnable)
	
	If Not acoDoor.fOpen Then
	'	Don't walk through door until we've opened it.
		Xnav.mstate = Xnav.mstateStop
		
	'	No way (without assessing) to know a priori whether the door is locked, 
	'	so assume it's unlocked until proven otherwise.
		fLocked = False
		Do
			If fLocked Then
				Call TryUnlock()
			End If
			
			Call TryOpen()
			
		Loop Until acoDoor.fOpen
	End If
	
	Call TraceExit("WayptDoor.EnsureOpen", "")
End Function

Private Function TryUnlock()
	Call TraceEnter("WayptDoor.TryUnlock", "")
	
	Dim acoKey
	If szKey <> "" Then
		Set acoKey = skapi.AcoFindInInv(szKey)
	Else
		Set acoKey = Nothing
	End If
	If acoKey Is Nothing And skapi.SkinfoFromSkid(skidLockpick).lvlCur > 0 Then
	'	We have no key, but we do have lockpick skill; see if it's worth trying to pick the lock.
		Dim oai : Set oai = Assess.OaiFromAco(acoDoor, -1)
		If Not oai Is Nothing Then
			If oai.ibi.diffLock - skapi.SkinfoFromSkid(skidLockpick).lvlCur <= 10 Then
				Set acoKey = skapi.AcoFindInInv("Peerless Lockpick")
			End If
		End If
	End If
	If acoKey Is Nothing Then
		If szKey <> "" Then
			Call RaiseCmr("You don't have a " + szKey + ".")
		Else
			Call RaiseCmr(acoDoor.szName + " is locked.")
		End If
	End If
	
	Call skapi.AddHandler(evidOnSetFlagOther, Me)
	
'	Try to unlock the door.
	Call acoKey.UseOnAco(acoDoor)
	
'	Wait for the result.
	Dim timer : Set timer = skapi.TimerNew()
	Do While fLocked
		Call Pauser.Check()
		Call CheckTimeout(timer, 5, "WayptDoor.TryUnlock")
		Call skapi.WaitEvent(300, wemFullTimeout)
	Loop
	
	Call skapi.RemoveHandler(evidOnSetFlagOther, Me)
	
	Call Pauser.Check()
	
	Call TraceExit("WayptDoor.TryUnlock", "")
End Function

Private Function TryOpen()
	Call TraceEnter("WayptDoor.TryOpen", "")
	
	Call skapi.AddHandler(evidOnTipMessage, Me)
	
'	Try to open the door.
	Call acoDoor.Use()
	
'	Wait for the result.
	Dim timer : Set timer = skapi.TimerNew()
	Do Until acoDoor.fOpen Or fLocked
		Call Pauser.Check()
		Call CheckTimeout(timer, 5, "WayptDoor.TryOpen")
		Call skapi.WaitEvent(300, wemFullTimeout)
	Loop
	
	Call skapi.RemoveHandler(evidOnTipMessage, Me)
	
	Call Pauser.Check()
	
	Call TraceExit("WayptDoor.TryOpen", "")
End Function


End Class


' Instantiator function.
Public Function WayptDoorNew()
	Set WayptDoorNew = New WayptDoorCls
End Function

' Register the waypoint type.
Call Xnav.RegisterType("Door", "WayptDoorNew")


