Option Explicit
Class WayptIncludeCls
' Normal waypoint class.


'---------------------------------------
' Type-specific properties and methods:
'---------------------------------------

Public xroute			' The route to include.
Public fReverse			' True if xroute should be executed in reverse order.

' Initialize this waypoint from the given parameters.
Public Function Init(xrouteA, fReverseA)
	Call TraceEnterEx("WayptInclude.Init", SzTraceObj(xrouteA) + ", " + _
	  CStr(fReverseA), Xnav.tef)
	
	Set xroute = xrouteA
	fReverse = fReverseA
	
	Call TraceExit("WayptInclude.Init", "")
End Function


'----------------------------------
' Required properties and methods:
'----------------------------------

' The XML tag name of this waypoint type.
Public Property Get szTag()
	szTag = "Include"
End Property

' True if this leg should be walked, false if it should be run.
' Xnav handles initialization of this property, but each waypoint class must define it.
Public fWalk

' The maploc Xnav should steer toward.
Public Property Get maplocIn()
	If xroute Is Nothing Then
		Set maplocIn = Nothing
	ElseIf fReverse Then
		Set maplocIn = xroute.maplocLast
	Else
		Set maplocIn = xroute.maplocFirst
	End If
End Property

' The maploc we're at after the arrival action has taken place.
' For waypoints with no arrival action, this is the same as maplocIn.
Public Property Get maplocOut()
	If xroute Is Nothing Then
		Set maplocOut = Nothing
	ElseIf fReverse Then
		Set maplocOut = xroute.maplocFirst
	Else
		Set maplocOut = xroute.maplocLast
	End If
End Property

' The arrival distance of this waypoint in MU, i.e. how accurately we need to approach it 
' before performing the arrival action (if any) and passing on to the next waypoint.
Public Property Get distArrive()
	If xroute Is Nothing Then
		distArrive = Xnav.distMeter
	ElseIf xroute.cwaypt = 0 Then
		distArrive = Xnav.distMeter
	ElseIf fReverse Then
		distArrive = xroute.rgwaypt(xroute.cwaypt - 1).distArrive
	Else
		distArrive = xroute.rgwaypt(0).distArrive
	End If
End Property

' The estimated time in seconds needed for this waypoint's arrival action.
' For waypoints with no arrival action, this is zero.
' Used by xroute.csecTravel.
Public Property Get csecArrive
	If xroute Is Nothing Then
		csecArrive = 0
	Else
		csecArrive = xroute.csecTravel
	End If
End Property

' Perform this waypoint's arrival action.
Public Function Arrive()
	Call TraceEnterEx("WayptInclude.Arrive", SzTraceObj(xroute), Xnav.tef)
	
	If Not (xroute Is Nothing) Then
		Dim xrouteT : Set xrouteT = xroute
		If fReverse Then
			Set xrouteT = xrouteT.XrouteReverse()
			If xrouteT Is Nothing Then
				Call RaiseCmr("Route " + xroute.szName + " is not reversible.")
			End If
		End If
		Call xrouteT.Go()
	End If
	
	Call TraceExit("WayptInclude.Arrive", "")
End Function

' Generate a new waypoint representing the reversal of this waypoint.
' Used by xroute.XrouteReverse.
Public Function WayptReverse()
	Call TraceEnterEx("WayptInclude.WayptReverse", SzTraceObj(xroute), Xnav.tef)
	Do
	
	If Not (xroute Is Nothing) Then
		If xroute.XrouteReverse() Is Nothing Then
			Set WayptReverse = Nothing
			Exit Do
		End If
	End If
	
	Set WayptReverse = WayptIncludeNew()
	Call WayptReverse.Init(xroute, Not fReverse)
	
	Loop While False
	Call TraceExit("WayptInclude.WayptReverse", "")
End Function

' Initialize this waypoint from skapi.maplocCur and, if relevant, skapi.acoSelected.
' Used by XrouteEdit.
Public Function FromMaplocCur()
	Call TraceEnterEx("WayptInclude.FromMaplocCur", "", Xnav.tef)
	
	Call Init(Nothing, False)
	
	Call TraceExit("WayptInclude.FromMaplocCur", "")
End Function

' Initialize this waypoint from an XML element.
Public Function FromXml(elem)
	Call TraceEnterEx("WayptInclude.FromXml", SzQuote(elem.xml), Xnav.tef)
	
	Dim szName : szName = SzGetAttr(elem, "szName")
	Dim fReverse : fReverse = FGetAttr(elem, "fReverse", False)
	
	Dim xroute
	If szName = "" Then
		Set xroute = Nothing
	Else
		Set xroute = Xnav.XrouteFromXdocSzName(elem.ownerDocument, szName)
		If xroute Is Nothing Then
			Dim elem2 : Set elem2 = elem.ownerDocument.SelectSingleNode("//Xroute[@szName = " + _
			  SzQuote(szName) + "]")
			If elem2 Is Nothing Then
				Call RaiseCmr("There is no route named " + SzQuote(szName) + ".")
			End If
			Set xroute = Xnav.XrouteFromXml(elem2, False)
		End If
	End If
	
	Call Init(xroute, fReverse)
	
	Call TraceExit("WayptInclude.FromXml", "")
End Function

' Save type-specific waypoint parameters to the given XML element.
' If fVerbose is True, all waypoint parameters should be made explicit in the XML.  
' (This is needed by XrouteEdit.)  If fVerbose is False, parameters with default 
' values may be omitted.
Public Function ToXml(elem, fVerbose)
	Call TraceEnterEx("WayptInclude.ToXml", CStr(fVerbose), Xnav.tef)
	
	Dim szName
	If xroute Is Nothing Then
		szName = ""
	Else
		szName = xroute.szName
	End If
	Call SetSzAttr(elem, "szName", szName)
	
	If fReverse Or fVerbose Then
		Call SetFAttr(elem, "fReverse", fReverse)
	End If
	
	Call TraceExit("WayptInclude.ToXml", "")
End Function


'-----------
' Privates:
'-----------

Private Sub Class_Initialize()
	
	Set xroute = Nothing
	fReverse = False
	
End Sub


End Class


' Instantiator function.
Public Function WayptIncludeNew()
	Set WayptIncludeNew = New WayptIncludeCls
End Function

' Register the waypoint type.
Call Xnav.RegisterType("Include", "WayptIncludeNew")


