Option Explicit
Class WayptPortalCls
' Portal waypoint class.


'---------------------------------------
' Type-specific properties and methods:
'---------------------------------------

Public szName			' The name of this portal.
Public maplocIn			' The portal's location.
Public maplocOut		' The portal's destination.

' Initialize this waypoint from the given parameters.
Public Function Init(szNameA, maplocInA, maplocOutA)
	Call TraceEnterEx("WayptPortal.Init", SzQuote(szNameA) + ", " + _
	  SzTraceMaploc(maplocInA) + ", " + SzTraceMaploc(maplocOutA), Xnav.tef)
	
	szName = szNameA
	Set maplocIn = maplocInA
	Set maplocOut = maplocOutA
	If maplocOut Is Nothing Then
		Set maplocOut = skapi.MaplocFromLatLng(0, 0)
	End If
	
	Call TraceExit("WayptPortal.Init", "")
End Function


'----------------------------------
' Required properties and methods:
'----------------------------------

' The XML tag name of this waypoint type.
Public Property Get szTag()
	szTag = "Portal"
End Property

' True if this leg should be walked, false if it should be run.
' Xnav handles initialization of this property, but each waypoint class must define it.
Public fWalk

' maplocIn and maplocOut are listed under "Type-specific properties and 
' methods" above.

' The arrival distance of this waypoint in MU, i.e. how accurately we need to approach it 
' before performing the arrival action (if any) and passing on to the next waypoint.
' Give portal lots of leeway so we don't run through it by accident.
Public Property Get distArrive()
	distArrive = 20 * Xnav.distMeter
End Property

' The estimated time in seconds needed for this waypoint's arrival action.
' We get this from a running average kept in the registry.
Public Property Get csecArrive()
	csecArrive = regvarCsecArrive.val
End Property

' Perform this waypoint's arrival action.
Public Function Arrive()
	Call TraceEnterEx("WayptPortal.Arrive", SzQuote(szName), Xnav.tef)
	
'	Measure how long the portal travel takes.
	Dim timerElapsed : Set timerElapsed = skapi.TimerNew()
	
'	Locate the portal object.
	Dim acoPortal : Set acoPortal = AcoFindPortal()
	If acoPortal Is Nothing Then Call RaiseCmr("Can't find " + szName + ".")
	
'	Not necessary, but good visual feedback.
	Call skapi.SelectAco(acoPortal)
	
	Call skapi.AddHandler(evidOnTipMessage, Me)
	
	Dim timer : Set timer = skapi.TimerNew()
	Dim itry : itry = 0
	Do
	'	Use the portal.
		Call acoPortal.Use()
		
	'	Wait until we enter portal space.
		timer.cmsec = 0
		fFailed = False
		Do Until skapi.plig = pligInPortal Or fFailed
			Call Pauser.Check()
			If timer.cmsec > 20000 Then Exit Do
			Call skapi.WaitEvent(300, wemFullTimeout)
		Loop
		If skapi.plig = pligInPortal Then Exit Do
		
		itry = itry + 1
		If itry = 3 Then Call RaiseCmr("Could not enter portal " + szName + ".")
	Loop
	
	Call skapi.RemoveHandler(evidOnTipMessage, Me)
	
'	Wait for emergence from portal space.
	timer.cmsec = 0
	Do Until skapi.plig = pligInWorld
		Call Pauser.Check()
		Call CheckTimeout(timer, 60, "WayptPortal.Arrive")
		Call skapi.WaitEvent(300, wemFullTimeout)
	Loop
	
'	Fold the elapsed time into the running average time.
	regvarCsecArrive.val = regvarCsecArrive.val * 0.9 + timerElapsed.cmsec/1000 * 0.1
	
'	Update the destination coords if not already known.
	If maplocOut.lat = 0 And maplocOut.lng = 0 Then
		Set maplocOut = skapi.maplocCur
	ElseIf skapi.maplocCur.Dist2DToMaploc(maplocOut) > 0.3 Then
		Call RaiseCmr("Wrong portal destination.")
	End If
	
	Call TraceExit("WayptPortal.Arrive", "")
End Function

' Generate a new waypoint representing the reversal of this waypoint.
' Used by Xroute.XrouteReverse.
Public Function WayptReverse()
	Call TraceEnterEx("WayptPortal.WayptReverse", "", Xnav.tef)
	
'	Portaling is not reversible.
	Set WayptReverse = Nothing
	
	Call TraceExit("WayptPortal.WayptReverse", "")
End Function

' Initialize this waypoint from skapi.maplocCur and, if relevant, skapi.acoSelected.
' Used by XrouteEdit.
Public Function FromMaplocCur()
	Call TraceEnterEx("WayptPortal.FromMaplocCur", "", Xnav.tef)
	
'	Make sure we have a portal selected.
	Dim acoSel : Set acoSel = skapi.acoSelected
	If acoSel Is Nothing Then Call RaiseCmr("Please select a portal.")
	If Not FAcoIsPortal(acoSel) Then Call RaiseCmr("Please select a portal.")
	
	Call Init(acoSel.szName, acoSel.maploc, Assess.MaplocDstOfPortal(acoSel))
	
	Call TraceExit("WayptPortal.FromMaplocCur", "")
End Function

' Initialize this waypoint from an XML element.
Public Function FromXml(elem)
	Call TraceEnterEx("WayptPortal.FromXml", SzQuote(elem.xml), Xnav.tef)
	
	szName = SzGetAttr(elem, "szName")
	Set maplocIn = Xnav.MaplocOfElem(elem.SelectSingleNode("maplocIn"))
	Set maplocOut = Xnav.MaplocOfElem(elem.SelectSingleNode("maplocOut"))
	
	Call TraceExit("WayptPortal.FromXml", "")
End Function

' Save type-specific waypoint parameters to the given XML element.
' If fVerbose is True, all waypoint parameters should be made explicit in the XML.  
' (This is needed by XrouteEdit.)  If fVerbose is False, parameters with default 
' values may be omitted.
Public Function ToXml(elem, fVerbose)
	Call TraceEnterEx("WayptPortal.ToXml", CStr(fVerbose), Xnav.tef)
	
	Call SetSzAttr(elem, "szName", szName)
	
	Dim elemT : Set elemT = elem.ownerDocument.CreateElement("maplocIn")
	Call Xnav.SetElemMaploc(elemT, maplocIn, True, False)
	Call elem.AppendChild(elemT)
	
	Set elemT = elem.ownerDocument.CreateElement("maplocOut")
	Call Xnav.SetElemMaploc(elemT, maplocOut, True, False)
	Call elem.AppendChild(elemT)
	
	Call TraceExit("WayptPortal.ToXml", "")
End Function


'-----------------
' Event handlers:
'-----------------

Public Function OnTipMessage(szMsg)
	Call TraceEnterEx("WayptPortal.OnTipMessage", SzQuote(szMsg), Xnav.tef)
	
	If FEqSzI(szMsg, "Unable to move to object!") Then
		fFailed = True
	End If
	
	Call TraceExit("WayptPortal.OnTipMessage", "")
End Function


'----------------
' Unwind handler
'----------------

' Remove event handlers and such in case of error.
Public Function Unwind(erh)
	Call TraceEnterEx("WayptPortal.Unwind", "", tefEnable)
	
	Call skapi.RemoveHandler(evidOnTipMessage, Me)
	
	Call TraceExit("WayptPortal.Unwind", "")
End Function


'-----------
' Privates:
'-----------

Private regvarCsecArrive	' Registry variable containing the average transit time.
Private fFailed

Private Sub Class_Initialize()
	
	Set maplocIn = Nothing
	Set maplocOut = Nothing
	
'	Set up the registry variable used to keep the running average transit time.
	Set regvarCsecArrive = Registry.Regvar("Xnav\csecArrivePortal", 30)
	
End Sub

' Look for a portal with the expected name at the expected coords.
Private Function AcoFindPortal()
	Call TraceEnter("WayptPortal.AcoFindPortal", SzTraceMaploc(maplocIn))
	
	Dim acf : Set acf = skapi.AcfNew()
	Set acf.maploc = skapi.MaplocFromLatLng(maplocIn.lat, maplocIn.lng, _
	  skapi.maplocCur.z)
	acf.szName = szName
	acf.ocm = ocmPortal Or ocmNPC
	acf.distMax = 0.1
	
	Set AcoFindPortal = acf.AcoGetSorted(skapi.sortbyDist)
	If AcoFindPortal Is Nothing Then
	'	Not found; maybe we got the name wrong.  (Portal names in ACX Pro routes are 
	'	often wrong.)  Try again without it.
		acf.szName = ""
		Set AcoFindPortal = acf.AcoGetSorted(skapi.sortbyDist)
	End If
	
	Call TraceExit("WayptPortal.AcoFindPortal", SzTraceAco(AcoFindPortal))
End Function


End Class


' Instantiator function.
Public Function WayptPortalNew()
	Set WayptPortalNew = New WayptPortalCls
End Function

' Register the waypoint type.
Call Xnav.RegisterType("Portal", "WayptPortalNew")


