Option Explicit
Class WayptRecallCls
' Recall waypoint class (as in @house recall or @lifestone).

' Note that recall spells are not supported here, due to character-specific 
' complexities of component and mana management, fizzle detection, Item skill 
' buffs, etc.  However if you have spellcasting code that works for you, 
' it shouldn't be hard to use this class as a template for building your own 
' spell-based waypoint class.


'---------------------------------------
' Type-specific properties and methods:
'---------------------------------------

Public szRecall			' The specific recall command (@lifestone or whatever).
Public maplocIn			' The place to recall from.  This can be Nothing to 
						' recall from wherever you happen to be standing.
Public maplocOut		' The recall destination.

' Initialize this waypoint from the given parameters.
Public Function Init(szRecallA, maplocInA, maplocOutA)
	Call TraceEnterEx("WayptRecall.Init", SzQuote(szRecallA) + ", " + _
	  SzTraceMaploc(maplocInA) + ", " + SzTraceMaploc(maplocOutA), Xnav.tef)
	
	szRecall = szRecallA
	Set maplocIn = maplocInA
	Set maplocOut = maplocOutA
	
	Call TraceExit("WayptRecall.Init", "")
End Function


'----------------------------------
' Required properties and methods:
'----------------------------------

' The XML tag name of this waypoint type.
Public Property Get szTag()
	szTag = "Recall"
End Property

' True if this leg should be walked, false if it should be run.
' Xnav handles initialization of this property, but each waypoint class must define it.
Public fWalk

' maplocIn and maplocOut are listed under "Type-specific properties and 
' methods" above.

' The arrival distance of this waypoint in MU, i.e. how accurately we need to approach it 
' before performing the arrival action (if any) and passing on to the next waypoint.
Public Property Get distArrive()
	distArrive = 2 * Xnav.distMeter
End Property

' The estimated time in seconds needed for this waypoint's arrival action.
' We get this from a running average kept in the registry.
Public Property Get csecArrive()
	csecArrive = regvarCsecArrive.val
End Property

' Perform this waypoint's arrival action.
Public Function Arrive()
	Call TraceEnterEx("WayptRecall.Arrive", SzQuote(szRecall), Xnav.tef)
	
'	Measure how long the recall takes.
	Dim timer : Set timer = skapi.TimerNew()
	
'	Say the recall command (@lifestone or whatever).
	Call skapi.Keys("{return}")
	Call skapi.PasteSz(szRecall)
	Call skapi.Keys("{return}")
	
'	Wait for emergence from portal space.
	Call Xnav.WaitForPortal()
	
'	Fold the elapsed time into the running average time.
	regvarCsecArrive.val = regvarCsecArrive.val * 0.9 + timer.cmsec/1000 * 0.1
	
'	Update the destination coords if not already known.
	If maplocOut.lat = 0 And maplocOut.lng = 0 Then
		Set maplocOut = skapi.maplocCur
	ElseIf skapi.maplocCur.Dist2DToMaploc(maplocOut) > 0.3 Then
		Call RaiseCmr("Wrong recall destination.")
	End If
	
	Call TraceExit("WayptRecall.Arrive", "")
End Function

' Generate a new waypoint representing the reversal of this waypoint.
' Used by Xroute.XrouteReverse.
Public Function WayptReverse()
	Call TraceEnterEx("WayptRecall.WayptReverse", "", Xnav.tef)
	
'	Recalling is not reversible.
	Set WayptReverse = Nothing
	
	Call TraceExit("WayptRecall.WayptReverse", "")
End Function

' Initialize this waypoint from skapi.maplocCur and, if relevant, skapi.acoSelected.
' Used by XrouteEdit.
Public Function FromMaplocCur()
	Call TraceEnterEx("WayptRecall.FromMaplocCur", "", Xnav.tef)
	
	Call Init("@lifestone", Nothing, skapi.maplocCur)
	
	Call TraceExit("WayptRecall.FromMaplocCur", "")
End Function

' Initialize this waypoint from an XML element.
Public Function FromXml(elem)
	Call TraceEnterEx("WayptRecall.FromXml", SzQuote(elem.xml), Xnav.tef)
	
	szRecall = SzGetAttr(elem, "szRecall")
	
	Dim elemT : Set elemT = elem.SelectSingleNode("maplocIn")
	If elemT Is Nothing Then
		Set maplocIn = Nothing
	Else
		Set maplocIn = Xnav.MaplocOfElem(elemT)
	End If
	
	Set maplocOut = Xnav.MaplocOfElem(elem.SelectSingleNode("maplocOut"))
	
	Call TraceExit("WayptRecall.FromXml", "")
End Function

' Save type-specific waypoint parameters to the given XML element.
' If fVerbose is True, all waypoint parameters should be made explicit in the XML.  
' (This is needed by XrouteEdit.)  If fVerbose is False, parameters with default 
' values may be omitted.
Public Function ToXml(elem, fVerbose)
	Call TraceEnterEx("WayptRecall.ToXml", CStr(fVerbose), Xnav.tef)
	
	Call SetSzAttr(elem, "szRecall", szRecall)
	
	Dim elemT
	If Not (maplocIn Is Nothing) Then
		Set elemT = elem.ownerDocument.CreateElement("maplocIn")
		Call Xnav.SetElemMaploc(elemT, maplocIn, True, False)
		Call elem.AppendChild(elemT)
	End If
	
	Set elemT = elem.ownerDocument.CreateElement("maplocOut")
	Call Xnav.SetElemMaploc(elemT, maplocOut, True, False)
	Call elem.AppendChild(elemT)
	
	Call TraceExit("WayptRecall.ToXml", "")
End Function


'-----------
' Privates:
'-----------

Private regvarCsecArrive	' Registry variable containing the average transit time.

Private Sub Class_Initialize()
	
	Set maplocIn = Nothing
	Set maplocOut = Nothing
	
'	Set up the registry variable used to keep the running average transit time.
	Set regvarCsecArrive = Registry.Regvar("Xnav\csecArriveRecall", 35)
	
End Sub


End Class


' Instantiator function.
Public Function WayptRecallNew()
	Set WayptRecallNew = New WayptRecallCls
End Function

' Register the waypoint type.
Call Xnav.RegisterType("Recall", "WayptRecallNew")


