Option Explicit

' XML utility functions.


'------------------
' Public functions
'------------------

Public Const NODE_ELEMENT = 1
Public Const NODE_TEXT = 3

' Create and return a new XML document.  If szElemRoot is non-empty, create a 
' root element with that tag.  Otherwise leave the root element null.
Public Function XdocNew(szElemRoot)
	Call TraceEnter("XdocNew", SzQuote(szElemRoot))
	
	Dim xdoc : Set xdoc = CreateObject("Msxml2.DOMDocument.4.0")
	xdoc.async = False
	xdoc.preserveWhiteSpace = True
	Call xdoc.SetProperty("SelectionLanguage", "XPath")
	
	If szElemRoot <> "" Then
		Set xdoc.documentElement = xdoc.CreateElement(szElemRoot)
	End If
	
	Set XdocNew = xdoc
	
	Call TraceExit("XdocNew", SzTraceXdoc(XdocNew))
End Function

' Create and return an XML document from the specified XML file.
Public Function XdocLoadFile(szFile)
	Call TraceEnter("XdocLoadFile", SzQuote(szFile))
	
	Dim xdoc : Set xdoc = XdocNew("")
	
	Call xdoc.Load(szFile)
	
	Set XdocLoadFile = xdoc
	
	Call TraceExit("XdocLoadFile", SzTraceXdoc(XdocLoadFile))
End Function

' Create and return an XML document from the specified XML text.
Public Function XdocFromSz(szXml)
	Call TraceEnter("XdocFromSz", SzQuote(szXml))
	
	Dim xdoc : Set xdoc = XdocNew("")
	
	Call xdoc.LoadXml(szXml)
	
	Set XdocFromSz = xdoc
	
	Call TraceExit("XdocFromSz", SzTraceXdoc(XdocFromSz))
End Function

' Insert and delete white-space text nodes as needed to format the given XML 
' element and its descendants with human-readable indentation.  clevel is the 
' indentation level to start from (typically zero for the root element of a 
' document).
Public Sub AdjustElemIndent(elem, clevel)
	
	Dim fAnyChildren : fAnyChildren = False
	Dim szText : szText = ""
	
	Dim nodeChild, nodeNext
	Set nodeChild = elem.firstChild
	Do Until nodeChild Is Nothing
		Set nodeNext = nodeChild.nextSibling
		Select Case nodeChild.nodeType
		Case NODE_TEXT
		'	Remove existing text nodes.
			szText = szText + SzTrimWhiteSpace(nodeChild.nodeValue)
			Call elem.RemoveChild(nodeChild)
		Case NODE_ELEMENT
		'	Insert indentation.
			Call elem.InsertBefore( _
			  elem.ownerDocument.CreateTextNode(vbCrlf + Space((clevel + 1) * 4)), nodeChild)
		'	Recurse to fix up child elements.
			Call AdjustElemIndent(nodeChild, clevel + 1)
			fAnyChildren = True
		End Select
		Set nodeChild = nodeNext
	Loop
	
	If fAnyChildren Then
		Call elem.AppendChild(elem.ownerDocument.CreateTextNode(vbCrlf + Space(clevel * 4)))
	End If
	
	If szText <> "" Then
		Call elem.insertBefore(elem.ownerDocument.CreateTextNode(szText), elem.firstChild)
	End If
	
End Sub

' Return a shorthand string representation of the given XML document suitable 
' for use with TraceEnter/TraceExit.
Public Function SzTraceXdoc(xdoc)
	
	If xdoc Is Nothing Then
		SzTraceXdoc = "Nothing"
	Else
		SzTraceXdoc = SzTraceElem(xdoc.documentElement)
	End If
	
End Function

' Return a shorthand string representation of the given XML element suitable 
' for use with TraceEnter/TraceExit.
Public Function SzTraceElem(elem)
	
	If elem Is Nothing Then
		SzTraceElem = "Nothing"
	Else
		SzTraceElem = SzQuote(elem.tagName)
	End If
	
End Function


' Read and return the value of a Boolean XML attribute.  elem is the element 
' to read from, szAttr is the name of the attribute, and fDefault is the 
' default value (True or False) to return if the attribute is not present.
Public Function FGetAttr(elem, szAttr, fDefault)
	
	Dim vAttr : vAttr = elem.GetAttribute(szAttr)
	If IsNull(vAttr) Then
		FGetAttr = fDefault
	Else
		FGetAttr = CBool(vAttr)
	End If
	
End Function

' Set a Boolean attribute on the given XML element.  szAttr is the name of the 
' attribute, and f is its boolean value.
Public Function SetFAttr(elem, szAttr, f)
	
	Call elem.SetAttribute(szAttr, CStr(CBool(f)))
	
End Function


' Read and return the value of an XML attribute as a Long.  elem is the 
' element to read from, szAttr is the name of the attribute, and lDefault 
' is the default value to return if the attribute is not present.
Public Function LGetAttr(elem, szAttr, lDefault)
	
	Dim vAttr : vAttr = elem.GetAttribute(szAttr)
	If IsNull(vAttr) Then
		LGetAttr = lDefault
	Else
		LGetAttr = CLng(vAttr)
	End If
	
End Function

' Set a Long attribute on the given XML element.  szAttr is the name of the 
' attribute, and l is its Long value.
Public Function SetLAttr(elem, szAttr, l)
	
	Call elem.SetAttribute(szAttr, CLng(l))
	
End Function


' Read and return the value of an XML attribute as a Long.  elem is the 
' element to read from, szAttr is the name of the attribute, and lDefault 
' is the default value to return if the attribute is not present.
Public Function LGetAttrHex(elem, szAttr, lDefault)
	
	Dim vAttr : vAttr = elem.GetAttribute(szAttr)
	If IsNull(vAttr) Then
		LGetAttrHex = lDefault
	Else
		Dim szT : szT = vAttr
		If FPrefixMatchSz(szT, "0x") Then
			szT = Right(szT, Len(szT) - 2)
		End If
		LGetAttrHex = CLng("&H" + szT)
	End If
	
End Function

' Set a Long attribute on the given XML element.  szAttr is the name of the 
' attribute, and l is its Long value.
Public Function SetLAttrHex(elem, szAttr, l)
	
	Call elem.SetAttribute(szAttr, SzHex(l, 8))
	
End Function


' Read and return the value of an XML attribute as a single-precision 
' float.  elem is the element to read from, szAttr is the name of the 
' attribute, and floatDefault is the default value to return if the attribute 
' is not present.
Public Function FloatGetAttr(elem, szAttr, floatDefault)
	
	Dim vAttr : vAttr = elem.GetAttribute(szAttr)
	If IsNull(vAttr) Then
		FloatGetAttr = floatDefault
	Else
		If szDecimal <> "." Then
		'	We're in some locale where dot is not the decimal separator.
		'	Replace dot with the correct separator before converting.
			vAttr = SzReplace(vAttr, "\.", szDecimal)
		End If
		FloatGetAttr = CSng(vAttr)
	End If
	
End Function

' Set a floating-point attribute on the given XML element.  szAttr is the 
' name of the attribute, and float is its value.
Public Function SetFloatAttr(elem, szAttr, float, cdigit)
	
	Call elem.SetAttribute(szAttr, Round(float, cdigit))
	
End Function


' Read and return the value of a textual XML attribute.  elem is the element 
' to read from, and szAttr is the name of the attribute.  If not present, 
' return the empty string.
Public Function SzGetAttr(elem, szAttr)
	
	Dim vAttr : vAttr = elem.GetAttribute(szAttr)
	If IsNull(vAttr) Then
		SzGetAttr = ""
	Else
		SzGetAttr = CStr(vAttr)
	End If
	
End Function

' Set a text-valued attribute on the given XML element.  szAttr is the name of 
' the attribute, and sz is its string value.
Public Function SetSzAttr(elem, szAttr, sz)
	
	Call elem.SetAttribute(szAttr, CStr(sz))
	
End Function


